 /*
  * Pidgin auto resize buddy list height plugin.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02111-1301  USA
  *
  *
  * --  Raoul Berger <contact@raoulito.info>, 12/2008 - 04/2009, ver. 0.6
  *
  */

#include "autoresize.h"

/* convenient macros */
#define force_update()\
  { GtkRequisition req;\
    gtk_widget_size_request (gtkblist->treeview, &req);	\
    g_signal_emit_by_name (G_OBJECT (gtkblist->treeview), \
			 "size-request", &req);\
  }

#define PLUGIN_EXTRA(plugin)\
  ((PluginExtraVars*)(plugin->extra))

static void buddy_list_update_from_bestheight(PurplePlugin * plugin)
{
    GtkWidget *top_window;
    gint y, real_height, pref_max_height;
    GdkGeometry hints;

    top_window = pidgin_blist_get_default_gtk_blist()->window;
    gtk_window_get_position(GTK_WINDOW(top_window), NULL, &y);

    pref_max_height =
	(gdk_screen_height() * purple_prefs_get_int(PREF_MAXHEIGHT)) / 100;

    real_height = MIN(pref_max_height, PLUGIN_EXTRA(plugin)->best_height);

    hints.min_height = real_height;
    hints.max_height = real_height;
    hints.min_width = 0;	/* 1 actually */
    hints.max_width = gdk_screen_width();

    /* update buddy list height constraints */
    gtk_window_set_geometry_hints(GTK_WINDOW(top_window),
				  top_window, &hints,
				  GDK_HINT_MIN_SIZE | GDK_HINT_MAX_SIZE);
    PLUGIN_EXTRA(plugin)->expect_conf_event = TRUE;
}

static void
on_blist_widget_size_request_cb(GtkWidget * child_widget,
				GtkRequisition * requisition,
				gpointer data)
{
    PurplePlugin *plugin;
    GtkNotebook *notebook;
    gint other_heights;
    GSList *l;
    gboolean is_updated;

    plugin = (PurplePlugin *) data;
    notebook =
	GTK_NOTEBOOK(pidgin_blist_get_default_gtk_blist()->notebook);
    is_updated = FALSE;
    other_heights = 0;

    for (l = PLUGIN_EXTRA(plugin)->assoc; l != NULL; l = l->next) {
	InstanceHdlHeight *assoc = (InstanceHdlHeight *) l->data;

	if (G_OBJECT(child_widget) == assoc->instance && !is_updated) {	/* addr comp */
	    is_updated = TRUE;
	    assoc->height = requisition->height;
	} else {
	    other_heights += assoc->height;
	}
    }
    if (gtk_notebook_get_current_page(notebook) == 1) {
	PLUGIN_EXTRA(plugin)->best_height =
	    requisition->height + other_heights;
	buddy_list_update_from_bestheight(plugin);
    }
}

#ifdef G_OS_WIN32
#define HBOX_POS 6
#else
#define HBOX_POS 5
#endif

static void
on_blist_headline_hbox_hide_cb(GtkWidget * headline_hbox, gpointer data)
{
    PidginBuddyList *gtkblist;
    PurplePlugin *plugin;
    GSList *l;
    InstanceHdlHeight *assoc;

    gtkblist = pidgin_blist_get_default_gtk_blist();
    plugin = (PurplePlugin *) data;
    l = g_slist_nth(PLUGIN_EXTRA(plugin)->assoc, HBOX_POS);

    assoc = (InstanceHdlHeight *) l->data;

    g_assert(G_OBJECT(headline_hbox) == assoc->instance);

    /* the widget size remain unchanged even if hidden,
       here we make our plugin believe its size is 0 */
    assoc->height = 0;
    force_update();
}

static void set_no_accounts_enabled_constraints(PurplePlugin * plugin,
						gboolean expect_conf)
{
    GtkWidget *top_window;
    GdkGeometry hints;

#ifdef G_OS_WIN32
    PLUGIN_EXTRA(plugin)->noconections = TRUE;
#endif

    top_window = pidgin_blist_get_default_gtk_blist()->window;

    hints.min_height = 405;	/* see pidgin_blist_init() def. in gtkblist.c */
    hints.max_height = hints.min_height;
    hints.min_width = 0;
    hints.max_width = gdk_screen_width();

    gtk_window_set_geometry_hints(GTK_WINDOW(top_window),
				  top_window,
				  &hints,
				  GDK_HINT_MIN_SIZE |
				  GDK_HINT_MAX_SIZE | GDK_HINT_USER_POS);
    PLUGIN_EXTRA(plugin)->expect_conf_event = expect_conf;
}

static void
on_blist_notebook_change_current_page(GtkNotebook * notebook,
				      GtkNotebookPage * page,
				      guint page_num, gpointer data)
{
    (void) notebook;
    (void) page;

    if (page_num == 0) {
	set_no_accounts_enabled_constraints((PurplePlugin *) data, TRUE);
    } else if (page_num == 1) {
	return;			/* nothing to do */
    } else {
	return;			/* shouldn't happen */
    }
}

static gint init_size(GtkWidget * child_widget, gboolean bypass_check)
{
    /* workaround for blist->headline_hbox */
    if (GTK_WIDGET_VISIBLE(child_widget) || bypass_check) {
	GtkRequisition req;
	gtk_widget_size_request(child_widget, &req);

	return req.height;

    } else {
	return 0;
    }
}

static
gboolean buddylist_window_configure_event_cb(GtkWidget * blistwin,
					     GdkEventConfigure * event,
					     PurplePlugin * plugin)
{
    gint height, y, x;

    gtk_window_get_size(GTK_WINDOW(blistwin), NULL, &height);
    gtk_window_get_position(GTK_WINDOW(blistwin), &x, &y);

    /*
       NOTE :
       As some window managers (WMs) ignore or don't handle well the "gravity" property,
       we have to move the window when the anchor region is set to ANCHOR_BOTTOM.

       The (short) delay between gtk_window_set_geometry_hints() and gtk_window_move() calls may
       cause the window to appear at wrong position (hint is set while move isn't) during this
       delay.

       For some WMs like Metacity or Kwin, this is useless and setting "gravity" to
       SOUTH_WEST is enough and better. %FIXME: use m4 macros to detect current WM ? %
     */
    if ((AutoresizeAnchorType)
	purple_prefs_get_int(PREF_ANCHOR) == ANCHOR_BOTTOM &&
	PLUGIN_EXTRA(plugin)->expect_conf_event == TRUE &&
	PLUGIN_EXTRA(plugin)->y_down != -1) {
	gint diff = PLUGIN_EXTRA(plugin)->y_down - (y + height);
	/* MAX : don't put titlebar offscreen */
	gtk_window_move(GTK_WINDOW(blistwin), x, MAX(y + diff, 0));


	PLUGIN_EXTRA(plugin)->expect_conf_event = FALSE;
	return FALSE;
    }
    PLUGIN_EXTRA(plugin)->y_down = y + height;
    return FALSE;
}

#ifdef G_OS_WIN32
static
gboolean buddylist_window_window_state_event_cb(GtkWidget * blistwin,
						GdkEventWindowState *
						event,
						PurplePlugin * plugin)
{
    if (event->type == 32) {	/* buddy list was just (de)iconified */
	PidginBuddyList *gtkblist;
	gtkblist = pidgin_blist_get_default_gtk_blist();


	if (PLUGIN_EXTRA(plugin)->noconections == FALSE) {
	    PLUGIN_EXTRA(plugin)->best_height = PLUGIN_EXTRA(plugin)->best_height + 1;	/* don't ask me why... */
	} else {
	    PLUGIN_EXTRA(plugin)->best_height = 406;
	    PLUGIN_EXTRA(plugin)->noconections = FALSE;
	}

	buddy_list_update_from_bestheight(plugin);
	gdk_window_process_all_updates();
    }
    return TRUE;
}
#endif

static void attach_signal(PurplePlugin * plugin,
			  GtkWidget * child_widget,
			  const gchar * detailed_signal,
			  GCallback c_handler,
			  gboolean is_sensitive, gboolean bypass_check)
{
    InstanceHdlHeight *assoc;

    g_return_if_fail(child_widget != NULL);
    assoc = g_malloc0(sizeof(InstanceHdlHeight));

    /* remember height */
    if (is_sensitive) {
	assoc->height = init_size(child_widget, bypass_check);
    } else {
	assoc->height = 0;
    }

    /* attach signal to update buddy list height */
    if (detailed_signal && c_handler) {
	assoc->instance = G_OBJECT(child_widget);
	assoc->handler = g_signal_connect(G_OBJECT(child_widget),
					  detailed_signal,
					  c_handler, plugin);
    }
    PLUGIN_EXTRA(plugin)->assoc =
	g_slist_append(PLUGIN_EXTRA(plugin)->assoc, assoc);
}

static void attach_signals(PurplePlugin * plugin)
{
    PidginBuddyList *gtkblist = pidgin_blist_get_default_gtk_blist();

    attach_signal(plugin,
		  gtkblist->window,
		  "configure-event",
		  G_CALLBACK(buddylist_window_configure_event_cb),
		  FALSE, TRUE);

#ifdef G_OS_WIN32
    attach_signal(plugin,
		  gtkblist->window,
		  "window-state-event",
		  G_CALLBACK(buddylist_window_window_state_event_cb),
		  FALSE, TRUE);
#endif

    /* we need to track height requisitions from widgets packed
       inside the buddy list */
    attach_signal(plugin,
		  gtkblist->treeview, "size-request",
		  G_CALLBACK(on_blist_widget_size_request_cb),
		  TRUE, FALSE);
    attach_signal(plugin,
		  gtkblist->statusbox, "size-request",
		  G_CALLBACK(on_blist_widget_size_request_cb),
		  TRUE, FALSE);
    attach_signal(plugin,
		  gtkblist->error_buttons, "size-request",
		  G_CALLBACK(on_blist_widget_size_request_cb),
		  TRUE, FALSE);
    attach_signal(plugin,
		  gtkblist->scrollbook, "size-request",
		  G_CALLBACK(on_blist_widget_size_request_cb),
		  TRUE, FALSE);
    attach_signal(plugin,
		  gtkblist->headline_hbox, "size-request",
		  G_CALLBACK(on_blist_widget_size_request_cb),
		  TRUE, FALSE);
    attach_signal(plugin,
		  gtk_item_factory_get_widget
		  (gtkblist->ift, "<PurpleMain>"), "size-request",
		  G_CALLBACK(on_blist_widget_size_request_cb),
		  TRUE, FALSE);

    attach_signal(plugin,
		  gtkblist->notebook, "switch-page",
		  G_CALLBACK(on_blist_notebook_change_current_page),
		  FALSE, FALSE);
    attach_signal(plugin,
		  gtkblist->headline_hbox, "hide",
		  G_CALLBACK(on_blist_headline_hbox_hide_cb),
		  FALSE, FALSE);
    /* "hidden" separator is present in the buddy list, its height
       won't change but its size is important */
    attach_signal(plugin, gtk_hseparator_new(), NULL, NULL, TRUE, TRUE);
}

#if 0
static void change_anchor_aux(AutoresizeAnchorType anchor)
{
    GtkWidget *top_window = pidgin_blist_get_default_gtk_blist()->window;

    switch (anchor) {
    case ANCHOR_TITLEBAR:
	{
	    gtk_window_set_gravity(GTK_WINDOW(top_window),
				   GDK_GRAVITY_NORTH_WEST);
	    break;
	}
    case ANCHOR_BOTTOM:
	{
	    gtk_window_set_gravity(GTK_WINDOW(top_window),
				   GDK_GRAVITY_SOUTH_EAST);

	    break;
	}
    default:			/* shouldn't happen */
	return;
    }
}

static void
on_pref_anchor_changed(const char *name, PurplePrefType type,
		       gconstpointer val, gpointer data)
{
    change_anchor_aux((AutoresizeAnchorType) GPOINTER_TO_INT(val));
    (void) name;
    (void) type;
    (void) data;
}
#endif

static void plugin_load_cb(PurpleBuddyList * dummy, gpointer data)
{
    PurplePlugin *plugin = (PurplePlugin *) data;

#if 0
    change_anchor_aux((AutoresizeAnchorType)
		      purple_prefs_get_int(PREF_ANCHOR));
#endif
    attach_signals(plugin);

    if (purple_connections_get_all() == NULL)
#ifdef G_OS_WIN32
	set_no_accounts_enabled_constraints(plugin, FALSE);	/* win32 issue tmp fix */
#else
	set_no_accounts_enabled_constraints(plugin, TRUE);
#endif
    (void) dummy;
}

static gboolean plugin_load(PurplePlugin * plugin)
{
    PidginBuddyList *gtkblist;

    gtkblist = pidgin_blist_get_default_gtk_blist();
    plugin->extra = g_malloc0(sizeof(PluginExtraVars));

    PLUGIN_EXTRA(plugin)->y_down = -1;
#ifdef G_OS_WIN32
    PLUGIN_EXTRA(plugin)->noconections = TRUE;
#endif
    /* wait for the buddy list to be created */
    if (!
	(gtkblist && gtkblist->treeview && gtkblist->statusbox
	 && gtkblist->error_buttons)) {
	purple_signal_connect(pidgin_blist_get_handle(),
			      "gtkblist-created", gtkblist,
			      PURPLE_CALLBACK(plugin_load_cb), plugin);
    } else {			/* buddy list already created */
	plugin_load_cb(NULL, plugin);
	force_update();
    }

    (void) plugin;
    return TRUE;
}

static void detach_signals(gpointer data, gpointer user_data)
{
    InstanceHdlHeight *assoc = (InstanceHdlHeight *) data;

    if (assoc->instance)
	g_signal_handler_disconnect(assoc->instance, assoc->handler);
    g_free(assoc);
    (void) user_data;
}

static gboolean plugin_unload(PurplePlugin * plugin)
{
    PidginBuddyList *gtkblist = pidgin_blist_get_default_gtk_blist();
    GtkWidget *top_window = gtkblist->window;
    GdkGeometry hints;

    g_slist_foreach(PLUGIN_EXTRA(plugin)->assoc, detach_signals, NULL);
    g_slist_free(PLUGIN_EXTRA(plugin)->assoc);
    g_free(PLUGIN_EXTRA(plugin));

    hints.min_height = 0;	/* 1 actually */
    hints.max_height = gdk_screen_height();
    hints.min_width = 0;	/* 1 actually */
    hints.max_width = gdk_screen_width();

    /* remove constraints */
    gtk_window_set_geometry_hints(GTK_WINDOW(top_window),
				  top_window,
				  &hints,
				  GDK_HINT_MIN_SIZE | GDK_HINT_MAX_SIZE);
    return TRUE;
}

static void
on_pref_height_changed(const char *name, PurplePrefType type,
		       gconstpointer val, gpointer data)
{
    PurplePlugin *plugin = (PurplePlugin *) data;

    if (pidgin_blist_get_default_gtk_blist() == NULL ||
	gtk_notebook_get_current_page(GTK_NOTEBOOK
				      (pidgin_blist_get_default_gtk_blist
				       ()->notebook)) == 0)
	return;			/* no account is enabled */
    else
	buddy_list_update_from_bestheight(plugin);
    (void) name;
    (void) type;
    (void) val;
    (void) data;
}

static void disconnect_prefs_callbacks(GtkObject * object, gpointer data)
{
    PurplePlugin *plugin = (PurplePlugin *) data;
    purple_prefs_disconnect_by_handle(plugin);



    (void) object;
}

static GtkWidget *get_config_frame(PurplePlugin * plugin)
{
    GtkWidget *ret;
    GtkWidget *frame;
    PurplePluginPrefFrame *pppf = purple_plugin_pref_frame_new();

    ret = gtk_vbox_new(FALSE, PIDGIN_HIG_CAT_SPACE);
    gtk_container_set_border_width(GTK_CONTAINER(ret), PIDGIN_HIG_BORDER);
    /* install prefs callbacks */
    g_signal_connect(GTK_OBJECT(ret), "destroy",
		     G_CALLBACK(disconnect_prefs_callbacks), plugin);
    purple_prefs_connect_callback(purple_prefs_get_handle(),
				  PREF_MAXHEIGHT,
				  on_pref_height_changed, plugin);
#if 0
    purple_prefs_connect_callback(purple_prefs_get_handle(),
				  PREF_ANCHOR, on_pref_anchor_changed,
				  NULL);
#endif
    purple_plugin_pref_frame_add(pppf,
				 purple_plugin_pref_new_with_label
				 (PREF_UI_FRAME1));
    frame = pidgin_plugin_pref_create_frame(pppf);

    gtk_box_pack_start(GTK_BOX(ret), frame, FALSE, FALSE, 0);

    /* set max % height widget */
    pidgin_prefs_labeled_spin_button(frame,
				     PREF_UI_MAXHEIGHT,
				     PREF_MAXHEIGHT, MIN_PTAGE_HEIGHT,
				     100, NULL);
    /* set anchor option widget */
    pidgin_prefs_dropdown(frame, PREF_UI_ANCHOR, PURPLE_PREF_INT,
			  PREF_ANCHOR,
			  PREF_ANCHOR_O1, ANCHOR_TITLEBAR,
			  PREF_ANCHOR_O2, ANCHOR_BOTTOM, NULL);

    gtk_widget_show_all(ret);
    return ret;
}

static PidginPluginUiInfo ui_info = {
    get_config_frame,
    0,				/* reserved */
    /* padding */
    NULL,
    NULL,
    NULL,
    NULL
};

static PurplePluginInfo info = {
    PURPLE_PLUGIN_MAGIC,
    PURPLE_MAJOR_VERSION,
    PURPLE_MINOR_VERSION,
    PURPLE_PLUGIN_STANDARD,
    PIDGIN_PLUGIN_TYPE,
    0,
    NULL,
    PURPLE_PRIORITY_DEFAULT,
    PLUGIN_ID,
    NULL,
    PLUGIN_VERSION,
    NULL,
    NULL,
    PLUGIN_AUTHOR,
    PLUGIN_WEBSITE,

    plugin_load,
    plugin_unload,
    NULL,
    &ui_info,
    NULL,
    NULL,
    NULL,
    /* padding */
    NULL,
    NULL,
    NULL,
    NULL
};

static void init_plugin(PurplePlugin * plugin)
{
#ifdef ENABLE_NLS
    bindtextdomain(GETTEXT_PACKAGE, LOCALEDIR);
    bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
#endif
    purple_prefs_add_none(PREF_PREFIX);

    /* default values, if node already created, nothing is done */
    purple_prefs_add_int(PREF_MAXHEIGHT, PREF_MAXHEIGHT_DEFAULT);
    purple_prefs_add_int(PREF_ANCHOR, ANCHOR_TITLEBAR);

    /* cleanup old options (ver < 0.5) */
    purple_prefs_remove(PREF_PREFIX "/y_bottom");
    purple_prefs_remove(PREF_PREFIX "/x_bottom");

    info.name = PLUGIN_NAME;
    info.summary = PLUGIN_SUMMARY;
    info.description = PLUGIN_DESCRIPTION;

    (void) plugin;
}

PURPLE_INIT_PLUGIN(auto_resize, init_plugin, info)

/* end of autoresize.c */
