/*
 * Pidgin auto resize buddy list height plugin - header -
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02111-1301  USA
 *
 *
 * --  Raoul Berger <contact@raoulito.info>, 04/2009, ver. 0.6
 *
 */


#ifdef HAVE_CONFIG_H
# include "../ar_config.h"
#endif

#define PURPLE_PLUGINS

#include <glib/gi18n-lib.h>
#include <gtkpluginpref.h>
#include <plugin.h>
#include <gtkprefs.h>
#include <version.h>
#include <gtkplugin.h>
#include <gtkblist.h>

/* various infos about this plugin */
#define PLUGIN_ID "gtk-rberger-auto_resize"

#define PLUGIN_NAME	_("Autoresize")
#define PLUGIN_SUMMARY	_("Auto-resize buddy list height.")
#define PLUGIN_DESCRIPTION _("This plugin will automatically resize your buddy " \
			     "list height to always fit its contents.")

#define PLUGIN_AUTHOR	"Raoul Berger <contact@raoulito.info>"
#define PLUGIN_WEBSITE 	"http://raoulito.info/pidgin_autoresize"
#define PLUGIN_VERSION  "0.6"

/* xml prefs keys */
#define	PREF_PREFIX	"/plugins/gtk/" PLUGIN_ID
#define PREF_MAXHEIGHT	PREF_PREFIX "/maxheight"
#define PREF_ANCHOR	PREF_PREFIX "/anchor"

/* prefs UI strings */
#define PREF_UI_FRAME1 _("-Configure-")
#define PREF_UI_MAXHEIGHT _("% of screen height:")
#define PREF_MAXHEIGHT_DEFAULT 85
#define MIN_PTAGE_HEIGHT 20	/* min hsize in % of screen height */
#define PREF_UI_ANCHOR _("Anchor region")
#define PREF_ANCHOR_O1 _("Titlebar")
#define PREF_ANCHOR_O2 _("Bottom")

typedef enum {
    ANCHOR_TITLEBAR,
    ANCHOR_BOTTOM,
} AutoresizeAnchorType;

/* assoc struct */
typedef struct {
    gpointer instance;
    gint handler;		/* keep signal handler id to detach instance */
    gint height;		/* current height of widget (instance) */
} InstanceHdlHeight;

/* vars needed by plugin */
typedef struct {
    GSList *assoc;		/* list of InstanceHdlHeight */
    gint best_height;		/* buddy list current perfect height */
    gint y_down;
    gboolean expect_conf_event;	/* TRUE if we are waiting for a  "configure_event" event */
#ifdef G_OS_WIN32
    gboolean noconections;
#endif
} PluginExtraVars;

/* end of autoresize.h */
